; source re-entered 1/4/2010 by Lee Hart
; (leeahart at earthlink.net) with
; special thanks to Richard, Ray Sills,
; William McKee, Chris Elmquist, Steven
; Valin, C.Wardell, and Doug Jackson.
; Copyright 2010 by Lee Hart
;
;---------------------------------------
;
;	 *****  *******  *     *
;	*     *    *     *     *
;	*     *    *     *     *
;	 *****     *     *******
;	*     *    *     *     *
;	*     *    *     *     *
;	 *****     *     *     *
;
;---------------------------------------
;
; 8TH is a programming system for the
; RCA "COSMAC" 1802 microprocessor. It
; provides a high-level block structured
; extensible language that combines the
; functions of a monitor, editor, immediate
; mode interpreter, compiler, and assembler
; in a single unified package. It is
; optimized to produce ROMable software
; for control systems and dedicated
; applications of the 1802.
;
; 8TH is essentially an 8-bit version of
; the FORTH language, and offers faster
; program execution with extremely low
; memory requirements. It presents a
; friendly, interactive programming
; environment with logical consistent
; rules throughout.
;
; written by Lee A. Hart
;   Technical Micro Systems, Inc.
;   366 Cloverdale
;   Ann Arbor, MI 48105
;   (313) 994-0784
;
; original copyright 9/2/81
;
; serial in, serial out version

;---------------------------------------
;	Definitions
;---------------------------------------
;
; WORD: Any string of ASCII characters
;       (up to 255). Words can either be
;	manipulated as data, or used as
;       names in 8TH.
;
; NAME: A word used to identify a par-
;	ticular program, subroutine, var-
;	iable, etc. Every structure in
;	8TH has a name.
;
; DICTIONARY: A list of all the names
;	currently in 8TH and their defi-
;	nitions. Definitions can be in
;	executable machine code, in 8TH
;	(defined in terms of other words)
;	or just a block of data. The
;	dictionary grows as new words are
;	added, and contracts as words are
;	deleted.
;
; VOCABULARY: The dictionary is divided
;	into a number of vocabularies,
;	each containing the words related
;	to a particular task. for example,
;	the "8TH" vocabulary contains all
;	the words recognized by 8TH itself.
;	8TH will recognize words in either
;	of two vocabularies at a time
;	(designated by the variables
;	"context" and "current").
;
; CURRENT vocabulary: The vocabulary
;	currently being created, where
;	words are added or deleted. When
;	words are typed in, the current
;	vocabulary is searched first.
;
; CONTEXT vocabulary: The vocabulary
;	of the program being executed
;	(normally 8TH itself). When
;	words are typed in, the context
;	vocabulary is searched second.
;
; 8TH:	The name of the vocabulary
;	containing all of the words that
;	make up 8TH itself. Initially,
;	"8TH" is the context vocabulary.
;
; COMPILER: The name of the vocabulary
;	containing all of the words used
;	to compile new definitions into
;	the dictionary. This vocabulary
;	is used by the ":" word.
;
; USER: The name of an empty vocabulary
;	available for the user of 8TH.
;	"User" is initially the current
;	vocabulary, so all new
;	definitions will be added to it.
;
; DATA STACK (or just STACK): A last-in
;	first-out stack used to pass
;	data between 8TH words. Data can
;	in general be constants, strings,
;	variables, addresses, etc.
;
; TOS:	Top byte (8 bits) on data stack.
;
; NOS:	Next byte on data stack.
;
; 3OS:	3rd byte on data stack.
;
; 4OS:	4th byte on data stack.
;
; RETURN STACK: A last-in first-out
;	stack generally used for storing
;	return addresses & loop counters.
;
; TORS: Top byte (8 bits) of the return
;	stack.
;
; NORS: Next byte on the return stack.
;
;---------------------------------------
;	Programming Conventions
;---------------------------------------
;
;  1.	The data stack moves downward as
;	data is pushed onto it. The data
;	stack pointer is left pointing to
;	the next unused byte.
;
;  2.	The return stack moves upward as
;	data is pushed onto it. The return
;	stack pointer is left pointing to
;	the next unused byte.
;
;  3.	Multibyte numbers always have the
;	most significant byte at the lowest
;	address, and the least significant
;	byte in the highest address.
;
;  4.	Strings have a length byte (from
;	0 to 255) at the lowest address.
;	The leftmost character is at the
;	next higher address, etc. and the
;	rightmost character is at the
;	highest address. The null string
;	(no elements) is a single byte
;	containing a "0". Characters within
;	a string can be any 8-bit value.
;
;---------------------------------------
;	Register Assignments
;---------------------------------------
;
;
;    REGISTER	DESCRIPTION
;    --------	-----------
dmar:	equ 0	; DMA pointer:
		; R0 is used briefly as the
		; program counter following
		; a hardware "clear". There-
		; after R0 is used only for
		; DMA.
		;
intr:	equ 1	; Interrupt Program Counter:
		; R1 is used only as a dedi-
		; cated program counter for
		; interrupt handlers.
		;
stack:	equ 2	; Data Stack Pointer:
		; R2 is the 8TH data stack
		; pointer. It must always be
		; pointing at either TOS or
		; the top free byte of RAM
		; (address of TOS-1).
		;
;	    3	; Machine Program Counter:
		; R3 is used as the program
		; counter for all machine
		; language programs, except
		; where otherwise indicated.
		;
call:	equ 4	; Call Program Counter:
		; R4 is the dedicated program
		; counter for the 8TH "call"
		; subroutine.
		;
next:	equ 5	; Inner Interpreter PC:
		; R5 is the dedicated program
		; counter for the 8TH "next"
		; inner interpreter. All
		; machine language routines
		; return to "next" via a
		; "SEP NEXT" instruction.
		;
rsp:	equ 6	; Return Stack Pointer:
		; R6 is the 8TH return stack
		; pointer. It must always be
		; pointing at TORS, or the
		; top free byte on the return
		; stack (address of TORS+1).
		;
ppc:	equ 7	; 8TH Pseudo Program Counter:
		; R7 points to the address
		; of the next word to be
		; executed by 8TH.
		;
variable: equ 8 ; Variable Program Counter:
		; R8 is the dedicated program
		; counter for pushing the
		; address of an 8TH variable
		; onto the data stack.
		;
vbase:	equ 9	; Variable Base Address:
		; R9 points to the upper end
		; of the block of RAM where
		; variables are stored. All
		; variables are accessed by
		; subtracting an offset from
		; this base address.
		;
;	   10	; RA Scratch Register:
;	   11	; RB Scratch Register:
		; RA and RB are used for
		; temporary storage within
		; 8TH machine language words.
		;
timer: equ 12	; RC Timer Register:
		; RC is decremented by the
		; interrupt handler every
		; 10 msec (approx.) and
		; can be used as a timer.
		;
;	   13	; RD: unused by 8TH
;	   14	; RE: unused by 8TH
;	   15	; RF: unused by 8TH
	page
;---------------------------------------
;	Initialization Words
;---------------------------------------
;
; clr:	Entry point into 8TH with P=X=0
;	following a hardware clear ("cold
;	start").
; clear0: Disables interrupts, changes the
;	program counter to register R3,
;	and continues into "clear".
;
	org 0		; origin of 8TH:
clr:	dis		;>disable interrupts
	db 0
	ghi r0		; initialize R3,
	phi r3		;   make it the
	ldi clear1	;   program counter,
	plo r3		;   and continue
	sep r3		;   at clear1:
;
; clear: The first 8TH word executed
;	(a "cold start"). It initializes
;	the 8TH pseudo-machine registers
;	and the context vocabulary
;	variables to the values indicated
;	in the initialization table (see
;	"initable"). Execution then begins
;	with the "1st" word (normally
;	"intrpret") with interrupts
;	disabled, the data and return
;	stacks cleared, and the current
;	vocabulary empty.
;
;	If continuous RAM is not present
;	from "minram" to "maxram", clear
;	will seek a continuous block of
;	RAM inside these limits. It then
;	corrects the 8TH CPU registers
;	and context vocabulary variables
;	to the limits of RAM actually
;	found. If insufficient RAM is
;	found, the "Q" output is set and
;	program execution is halted.
;
;	Clear works by setting the
;	variable base pointer (R9) to
;	"initable" and continuing into
;	"reset".
;
clear1: sex r3		;>disable interrupts
	dis
	db 23h		; set X=2, P=3
	ldi a.1(initable); set variable base
	phi vbase	;   to "initable"
	ldi a.0(initable)
	plo vbase	; continue to "reset"

; reset (warm start): Initializes the
;	8TH pseudo-machine CPU registers
;	and context vocabulary variables.
;	Execution then begins with the
;	"1st" word (usually "intrpret")
;	with interrupts disabled, the
;	data and return stacks empty, and
;	the current vocabulary empty.
;
;	"Reset" works by loading the
;	variables "maxram" and "minram"
;	(via the variable base pointer)
;	into the CPU, and then testing for
;	a continuous block of RAM between
;	these limits. it then corrects
;	the values of "maxram", "minram",
;	"context", "current", "user"
;	"sp0", and "rsp", and moves all
;	variables to the new RAM limits.
;	If insufficient RAM is found, the
;	"Q" output is set and execution
;	is halted. Reset then continues
;	into "abort".
;
reset1: sex r3		;>disable interrupts
	dis
	db 0a3h		; and set X=A, P=3
	ghi vbase	; set pointer=vbase
	phi ra
	glo vbase
	plo ra
	ldx		; set R8=maxram
	plo r8
	dec ra
	ldx
	phi r8
	dec ra
	glo r8		; set size =
	sm		;   maxram-minram
	plo rb
	dec ra
	ghi r8
	smb
	phi rb
	dec ra
	glo r8		; set #variables =
	sm		;   maxram-sp0
	plo r4
	plo r5		; set #var =
	dec ra		;   maxram-sp0
	ghi r8		;   (2 copies)
	smb		; set found=false
	phi r4		;   (DF=1)
	phi r5
	ldi 0		; set #bytes=0
	phi r6		;   (#bytes of RAM)
	plo r6
	sex r8

; testram: seeks a contiguous block of
;	RAM between "maxram" and "minram",
;	and initializes registers vbase
;	and origin accordingly.
;
			; repeat:
testram: ldn r8		; - load byte
	plo r7		; - save a copy
	xri 0ffh	; - complement it
	str r8		; - try storing it
	xor		; - see if it worked
	bz ram		;   (D=0 if RAM)
	bdf restore	; - case 1: not RAM
			;   and found=true
			;   (found minram)
	glo r7		; - get copy
	str r8		; - restore it
	br movevar	; - go move var.
			; - case 2: RAM
ram:	inc r6		; - inc. #bytes
	bnf restore	; - if found=false
	shl		;   set found=true
	ghi r8		;   set vbase=rs
	phi vbase
	glo r8		;   (found maxram)
	plo vbase
restore: glo r7		; - else restore RAM
	stxd		;   decr. R8
	glo rb
	bnz $+3
	ghi rb		; - decrement size
	dec rb		; - until size < 0
	bnz testram	; - halt if size <0
	bdf halt	;   and found=false

; movevar: Move variables from old location
;	to new location at "maxram".
;	(Note: Variables must not cross
;	the FFFF-0000 boundary)
;	  vbase = actual maxram
;	  r8 = actual maxram-1
;	  r6 = actual #bytes of RAM found
movevar:
	irx		; set R8 = minram
	ldn vbase	; free a byte of RAM
	plo r7		;   to allow math
	sex vbase
	inc ra		; set pointer =
	inc ra		;   old variable
	inc ra		;   location
	inc ra
	inc ra
	glo ra		; if vbase-pointer<0
	str vbase	;   then let up=no
	glo vbase	;   (block move
	sm		;   down, DF=0)
	ghi ra		; else let up=yes
	str vbase	;   (block move
	ghi vbase	;   up, DF=1)
	smb
	glo r7		; restore byte
	str vbase	;   to RAM
moveup:			; repeat:
	ldn ra		; - load via pointer
	dec ra		; - decr. pointer
	lsnf		; - if up=yes, then
	str vbase	;     store via vbase
	dec r6		;     decr. #bytes
	dec vbase	; - decr. vbase
	dec r4		; - decr. #variables
	ghi r4		; - until #variables=0
	bnz moveup
	glo r4
	bnz moveup
	ghi vbase	; set stack = vbase
	phi stack
	glo vbase
	plo stack
movedown:		; repeat:
	lda ra		; - load via pointer
			;   and incr. pointer
	lsdf		; - if up=no, then
	str vbase	;     store via vbase
	dec r6		;     decr. #bytes
	inc vbase	; - incr. vbase
	dec r5		; - decr. #var
	ghi r5		; - until #var=0
	bnz movedown
	glo r5
	bnz movedown

; halt: Set Q and halt if insufficient RAM
;	is found (8TH can't run)
;
	glo r6		; get #bytes of RAM
	smi 10		;   left after moving
	ghi r6		;   variables
	smbi 0
	lsdf		; if #bytes < 10,
halt:	seq		;   then set Q
	dec r3		;   and halt!

; Correct variables "maxram", "minram",
; "context", "current", "user", "sp0",
; and "rsp0" to actual values.

	glo vbase	; point RA
	smi 23		;   to "context"
	plo ra
	ghi vbase
	smbi 0
	phi ra
	glo vbase	; point RB to "8TH"
	smi 126
	plo rb
	ghi vbase
	smbi 0
	phi rb
	str ra		; point "context"
	inc ra		;   to "8TH"
	glo rb
	str ra
	inc rb		; point RB to "user"
	inc rb
	inc ra		; point RA to "current"
	ghi rb		;   and point "current"
	str ra		;   to "user"
	inc ra
	glo rb
	str ra
	inc rb		; point "user" to
	glo rb		;   empty user vocab
	smi 8
	str rb
	dec rb
	ghi rb
	smbi 0
	str rb
	ghi vbase	; set pointer = vbase
	phi ra
	glo vbase
	plo ra
	sex ra
	stxd		; set "maxram" = vbase
	ghi vbase
	stxd
	glo r8		; set "minram" = R8
	stxd
	ghi r8
	stxd
	glo stack	; set "sp0" = stack
	stxd
	ghi stack
	stxd
	glo r8		; set "rsp0" = R8
	stxd
	ghi r8
	stxd
; initialize remaining 8TH registers
	ldi 9		; pointer=pointer-10
	dec ra
	smi 1
	bnz $-3
	ldxa		; set variable(R8) =
	phi variable	;   "varat"
	ldxa
	plo variable
	ldxa		; set intr(R1):
	phi intr	;   "interrupt"
	ldxa
	plo intr
	ldxa		; set call(R4):
	phi call	;   "call"
	ldxa
	plo call
	ldxa		; set next(R5):
	phi next	;   "next"
	ldxa
	plo next
	lbr abort	;<go to abort

; next: Inner interpreter entered via
;	"SEP R5", "next" fetches the
;	address of the next 8TH word via
;	the pseudo-program counter and
;	executes it. If the address of
;	the next word begins with "0", it
;	is the special "load immediate"
;	operator, and the low byte is
;	immediately pushed onto the data 
;	stack. "Next" uses R5 as its
;	dedicated program counter, and so
;	must only be executed with P=R5.
;
	phi r3		; put in R3.1
	lda ppc		; load low byte
	plo r3		; put in R3.0
	sep r3		;<set PC to R3
next1:	lda ppc		;>load hi byte
	bnz $-5		; if load immediate,
	lda ppc		;   push low byte
	stxd		;   onto data stack,
	br next1	;   branch to next1

; call: Routine to call 8TH routines.
;	"Call" uses R4 as its dedicated
;	program counter.
;
	sep next	;<go to next
call1:	ghi ppc		;>push pseudo-
	str rsp		;   program counter
	inc rsp		;   to return stack
	glo ppc		;   (hi byte first)
	str rsp
	inc rsp
	glo r3		; set pseudo-program
	plo ppc		;   counter to addr
	ghi r3		;   spec'd by calling
	phi ppc		;   program (at R3)
	br $-11		; go to exit

; interrupt: Restarts a DMA scan of the 
;	keyboard and display, and decrements
;	"timer" (RC). Requires that the
;	CPU's interrupt line be connected to
;	the "row 9" (or "row 8" on the -X1)
;	BASYS/1 board). Interrupts will occur
;	every 11.266 msec with a 2.00 MHz
;	clock, and will require about 10% of
;	the CPU's execution time (2.562 msec
;	and 40% for -X1 board). "Interupt"
;	uses R1 as its dedicated program
;	counter.
;
	ret		;<exit
interupt: dec stack	;>save D on stack
	str stack
	dec timer	; decrement timer
	ldi 0e9h	; reset DMA pointer
	idl		;   (R8) after the 
	plo dmar	;   last DMA cycle
	out 1		; start DMA cycle
	dec stack
	ldx		; restore D
	sav		; push T and
	lbr $-12	; go to exit

;---------------------------------------
;	Beginning of normal 8TH words
;--------------------------------------- 
; clear (cold start): Clears the stacks,
;	dictionary, and variables to the
;	conditions when power was first
;	applied. Same as hardware "clear"
;	(see clear1).
;
clear:	lbr clear1
;
; reset (warm start): Clears the stacks
;	and resets the variables and
;	dictionary to the values indicated
;	by the context vocabulary
;	variables (see reset1).
;
reset:	lbr reset1
;
; abort (or 'Abort'): Disables interrupts,
;	clears the data and return stacks
;	& resumes execution at the outer
;	interpreter.
;     
;	"Abort" works by setting the 8TH
;	pseudo-program counter to the
;	word defined as "1st", and then
;	setting the data and return stack
;	pointers to "SP0" and "RSP0"
;	respectively.
;
abort:	sex r3		;>disable interrupts
	dis		;   (while moving
	db 23h		;    stack pointers)
	ghi r3		; change PC to RB
	phi rb
	glo r3
	plo rb
	inc rb
	sep rb
	glo vbase	; set pointer to
	smi 9		;   variable "1st"
	plo ra
	ghi vbase
	smbi 0
	phi ra
	lda ra		; set R3 = "1st"
	phi r3
	lda ra
	plo r3
	lda ra		; set RSP = "RSP0"
	phi rsp		;   (clears return)
	lda ra		;    stack)
	plo rsp
	lda ra		; set stack = "SP0"
	phi stack	;   (clears data
	lda ra		;    stack)
	plo stack
	sep r3		;<go to "1st"

; return: Routine to return from 8TH.
;
return: dec rsp		;>pop return address
	ldn rsp		;   from return stack
	plo ppc		;   and put it into
	dec rsp		;   pseudo-program
	ldn rsp		;   counter
	phi ppc		;   ...lo byte first
	sep next	;<go to next

; dup: Duplicates byte on top of stack.
;
dup:	irx		;>pop byte on
	ldx		;   top of stack
	stxd		; push it back on
	stxd		;   twice
	sep next	;<return to next

; ddrop: Drops top 2 bytes from stack.
;
ddrop:	irx		;>drop TOS
	irx		; drop NOS
	sep next	;<return to next

; drop: Drops top byte from the stack.
;
drop:	irx		;>drop TOS
	sep next	;<return to next

; over: Copy NOS and push it onto stack.
;
over:	ghi stack	;>copy stack pointer
	phi ra		;   into scratch
	glo stack	;   register RA
	plo ra
	inc ra		; increment it to
	inc ra		;   point to NOS
	ldn ra		; load NOS
	stxd		;   & push onto TOS
	sep next	;<return to next

; swap: Interchange TOS and NOS.
;
swap:	irx		;>pop byte on
	ldxa		;   top of stack
	phi ra		;   and save in RA.1
	ldx		; get next on stack
	plo ra		;   and save in RA.0
	ghi ra		; get old TOS And 
	stxd		;   push onto stack
	glo ra		; get old NOS and
	stxd		;   push onto stack
	sep next	;<return to next

; rot:	Rotate the top 3 bytes on stack
;	so TOS->NOS, NOS->3OS, 30S->TOS.
;
rot:	irx		;>pop byte on top
	ldxa		;   of stack and
	phi ra		;   save in RA.1
	ldxa		; pop next on stack
	plo ra		;   save in RA.0
	ldx		; get 3rd on stack
	phi rb		;   save in RB.1
	glo ra		; push old NOS from
	stxd		;   RA.0 onto stack
	ghi ra		; push old TOS from
	stxd		;   RA.1 onto stack
	ghi rb		; push old 3OS from
	stxd		;   RB.1 onto stack
	sep next	;<return to next

; + (add): Adds TOS and NOS.
;
add:	irx		;>pop top byte from
	ldxa		;   data stack
	add		; add TOS+NOS
	stxd		; push sum to stack
	sep next	;<return to next

; - (subtract): Subtract TOS from NOS.
;
subtr:	irx		;>pop top byte from
	ldxa		;   data stack
	sd		; subtract NOS-TOS
	stxd		; push result to TOS
	sep next	;<return to next

; and:	Bitwise "and" of TOS, NOS.
;
and1:	irx		;>pop top byte from
	ldxa		; data stack
	and		; TOS "and" NOS
	stxd		; push result to TOS
	sep next	;<return to next

; or:	Bitwise "or" of TOS, NOS.
;
or1:	irx		;>pop top byte from
	ldxa		;   data stack
	or		; TOS "or" NOS
	stxd		; push result to TOS
	sep next	;<return to next

; xor:	Bitwise "exclusive or" of TOS, NOS.
;
xor1:	irx		;>pop top byte from
	ldxa		; data stack
	xor		; TOS "xor" NOS
	stxd		; push results to TOS
	sep next	;<return to next

; not:	Pops top of stack and tests. If
;	TOS=false(zero), pushes true; if
;	TOS=true (non-zero), pushes false.
;
not1:	irx		;>pop byte on top of 
	ldx		;   data stack
	lsz		; if TOS=true
	ldi 0ffh	;   then load -1
	adi 1		; add 1; result is 0
			;   only if TOS=true
	stxd		; push result on TOS
	sep next	;<return to next

; = (equal): Test for NOS=TOS; returns boolean.
;
equal:	irx		;>pop top byte of 
	ldxa		;   stack
	xor		; compare TOS & NOS
			; D=00 if TOS=NOS
			; D= + otherwise
	sdi 0		; subtract D from 0
			; DF=1 if TOS=NOS
			; DF=0 otherwise
	ldi 0		; set D=0
	shlc		; shift DF into D0
	stxd		; push onto stack
	sep next	;<return to next

; >: Tests for NOS>TOS; returns boolean.
;
greater: irx		;>pop top byte
	ldxa		;   from stack
	sm		; subtract TOS-NOS
			; if NOS>TOS, DF=0
			; if NOS<=TOS, DF=1
	ldi 0ffh	; set D=ff
	adci 0		; add D(=ff)+0+DF
			; if NOS>TOS, D=FF(true)
			; if NOS<=TOS, D=0(false)
	stxd		; push results to TOS
	sep next	;< return to next

; <: Tests for NOS<TOS; returns boolean.
;
less:	irx		;>pop top byte of 
	ldxa		;   stack
	sd		; subtract TOS-NOS
			; if NOS<TOS, DF=0
			; if NOS>=TOS, DF=1
	ldi 0ffh	; set D=FF
	adci 0		; add D(=FF)+0+DF
			; if NOS<TOS, D=FF(true)
			; if NOS>=TOS, D=0(false)
	stxd		; push onto stack
	sep next	;<return to next

;-----------------------------------
;	Repeat-Until
;-----------------------------------

; (repeat): Pushes PPC onto return stack.
;
repeat: ghi ppc		;>pushes pseudo-
	str rsp		;   program counter
	inc rsp		;   to return stack;
	glo ppc		; hi byte first,
	str rsp		; lo byte second
	inc rsp
	sep next	;<return to next

; (until): Pop TOS and test; if false, go
;	back to address on return stack
;	(matching "repeat"). If true, pop
;	address on return stack and continue.
;
until:	irx		;>pop top byte from
	ldx		; data stack
	dec rsp
	lsz		; if not 0 (=true),           
	dec rsp		;   drop return addr
	sep next	;<  & return to next
	ldn rsp		; else pop return addr
	plo ppc		;   from return stack
	dec rsp		;   and load it into
	lda rsp		;   the pseudo-program
	phi ppc		;   counter
	glo ppc		; push PPC back on
	str rsp		;   to return stack
	inc rsp		;   for next pass
	sep next	;<return to next

;-----------------------------------
;	Do-Loop
;-----------------------------------
;
; (do): Begins a do-loop. "Do" expects 2
;	bytes on data stack: TOS=initial
;	value, and NOS=limiting value of
;	loop index counter. "Do" pushes
;	the PPC, then limit (NOS), then
;	initial value (TOS) onto return
;	stack for "loop" or "+loop" use.
;
do:	ghi ppc		;>push pseudo-program
	str rsp		;   counter onto
	inc rsp		;   return stack;
	glo ppc		;   hi byte first,
	str rsp		;   then lo byte.
	inc rsp
	irx		; pop TOS (initial
	ldxa		;   value of loop)
	plo ra		;   save temporarily
	ldx		; pop NOS (limit)
	str rsp		;   and push it onto
	inc rsp		;   return stack
	glo ra		; get initial value
	str rsp		;   and push it onto
	inc rsp		;   return stack
	sep next	;<return to next

; (loop): Expects 4 bytes on return stack;
;	TORS=index, NORS=limit, and 3ORS
;	and 4ORS=addr of the first word
;	of the do-loop. "Loop" adds 1 to 
;	the index and compares it to the
;	limit: If (index+1)>limit, the 4
;	bytes on return stack are dropped
;	and execution continues with the 
;	word following "loop". If (index
;	+1)=<limit, execution loops back
;	to the first word following the
;	previous "do", with all 4 bytes
;	on the return stack ready for the
;	next pass.
;(+loop): Same as loop, except TOS is 
;	added to index instead of 1.
;
loop:	ldi 1		;>load 1
	stxd		;   push onto TOS and
pluslp:			;   continue to +loop
	dec rsp		;>pop index from top
	ldn rsp		;   of return stack
	irx		; pop TOS and
	add		;   add TOS+index
	dec rsp		; point to limit
	bdf done	; if (TOS+index)>255
			;   then go to done
	str stack	; else save sum on TOS
	ldn rsp		; subtract
	sm		;   limit-(TOS+index)
	bnf done	; if <0, go to done
	glo rsp		; else set pointer
	plo ra		;   to top of return
	ghi rsp		;   stack (points at
	phi ra		;   limit)
	inc rsp		; point return stack
			;   pointer to index
	ldx		; push updated value
	str rsp		;   of index onto
	inc rsp		;   return stack
	dec ra		; set pointer to
	ldn ra		;   load addr.0
	plo ppc		;   and put in PPC.0
	dec ra		; set pointer to
	ldn ra		;   load addr.1
	phi PPC		;   and put in PPC.1
	sep next	;<return to next
done:	dec rsp		; drop addr.0
	dec rsp		; drop addr.1
	sep next	;<return to next

; i: Pushes current value of do-loop
;	index counter onto data stack.
i:	dec rsp		;>get copy of "i"
	lda rsp		;   from return stack
	stxd		; push onto TOS
	sep next	;<return to next

; exit: Used inside a do-loop to force
;	an exit at the next occurence of
;	"loop" by setting the loop limit 
;	to zero.
;
exit:	glo rsp		;>initialize pointer
	plo ra		;   to top of return
	ghi rsp		;   stack
	phi ra
	dec ra		; move pointer to 
	dec ra		;   point to limit
	ldi 0		; load 0 in
	str ra		;   place of limit
	sep next	;<return to next

;--------------------------------------
; Branch words (compiled If/Then/Else)
;--------------------------------------
;
; br0: Compiled "if", followed by a 2-byte
;	address. "BR0" pops TOS and tests:
;	If true, skip PPC over the address
;	and continue. If false, branch to
;	the specified address.
;
br0:	irx		;>pop top byte from
	ldx		;   data stack
	inc ppc
	lsz		; if TOS=true, then
	inc ppc		;   skip over addr
	sep next	;<  & return to next
	dec ppc		; else continue

; br: Compiled "else", followed by a
;	2-byte address. Unconditionally
;	branches to the specified address
;
br:	lda ppc		;>get branch addr.0
	str stack	;   and save on TOS
	ldn ppc		; get branch addr.1
	plo ppc		;   and put in PPC.0
	ldx		; get hi byte from TOS
	phi ppc		;   and put in PPC.1
	sep next	;<return to next

; ie: Pop TOS and set interrupt enable
;	status to its boolean value
;	(TOS=true=enable, TOS=false=disable).
;
ie:	irx		;>pop top of stack
	ldx
	sex 3
	lsz		; if TOS=true,
	ret		;   set IE=1
	db 23h
	lsnz		; if TOS=false,
	dis		;   set IE=0
	db 23h
	sep next	;<return to next

; ie@: Tests CPU interrupt enable (IE)
;	and pushes it onto TOS. Returns
;	TOS=true if interrupts enabled,
;	or TOS=false if disabled.
;
ieat:	ldi 1		;>assume status=true
	lsie		; if Ie=0, then
	ldi 0		;   status=false
	stxd		; push status
	sep next	;<return to next

; v@ (variable @): Entered via "Sep R8",
;	"v@" calculates the address of a
;	variable, and pushes it onto the 
;	stack for use by "@" or "!". "v@"
;	is followed by a 2-byte offset,
;	which is subtracted from the variable
;	base address (vbase) in R9.
;
	sep r5		;<return to next
varat:	inc r3		;>point R(x) to
	sex r3		;   offset.0
	glo vbase	; subtract
	sm		;   vbase.0-offset.0
	str stack	;   and push on TOS
	dec stack
	dec r3		; load offset.1
	ghi vbase	; subtract
	smb		;   vbase.1-offset.1
	sex stack
	stxd		;   and push on TOS
	br varat-1	; go to exit

; @ (at): Expects an address on the data
;	stack (TOS=addr.1), NOS=addr.0).
;	"@" pops the address, reads the
;	specified memory location, and
;	pushes its contents onto TOS.
;
at:	irx		;>initialize pointer
	ldxa		; to mem address:
	phi ra		; - addr.1
	ldx
	plo ra		; - addr.0
	ldn ra		; read via pointer
	stxd		; and push on tos
	sep next	;<return to next

; ! (excla): Expects 3 bytes on stack:
;	TOS=addr.1, NOS=addr.0, 3OS=data.
;	"!" pops all 3 bytes from the stack,
;	and writes the data byte into the
;	memory address specified.
;
excla:	irx		;>initialize pointer
	ldxa		; to mem. address:
	phi ra		; - addr.1
	ldxa
	plo ra		; - addr.0
	ldx		; pop data byte and
	str ra		;   store via pointer
	sep next	;<return to next

; emits: Types out one byte on top of
;	the stack via the serial output
;	device.
;
;	Serial output uses "Q": 1 start
;	bit, 8 data bits, and 1 stop bit.
;	First character is preceeded by a
;	start bit to aid synchronization.
;	The baud rate is determined by
;	the variable "baud".
;
emits:	glo vbase	;>initialize pointer
	smi 113		;   to the variable
	plo ra		;   "baud"
	ghi vbase
	smbi 0	
	phi ra
	ldn ra		; get "baud" and
	str rsp		;   store at TORS
	req		; begin a stop bit
	plo ra		; put baud in RA
stop1:	nop		; wait 1 bit time by
	glo ra		; counting down baud
	dec ra
	bnz stop1
start:			; output start bit
	irx		; pop character from
	ldx		;   top of stack
	seq		; begin a start bit
	plo rb		; save character
	ldi 10		; set #bits=10
	plo ra		; save #bits in RA.0
; output bits of character until #bits=0
bits:	ldn rsp		; get the variable
			; "baud" and:
	phi rb		; - save copy of it
	smi 1		; - loop for copy*4
	bdf $-2		;   machine cycles
	ghi rb		; - get copy of baud
	smi 35		; - subtract 35
	bdf $-8		; - repeat until "-"
	shl		; set DF=1
	dec ra		; decrement #bits
	glo rb		; shift next bit of
	rshr		;   character into DF
	plo rb		;   and DF into last
			;   bit of character
	lsdf		; output DF:
	seq		; if DF=0, ouput
	lskp	 	;   a "space"
	req		; if DF=1, output 
	nop		;   a "mark"
	glo ra		; get #bits to go
	bnz bits	; repeat until 0
	sep next	;<return to next

; keys: Gets 1 character from the serial
;	input and returns its ASCII code	
;	on TOS. If no valid key is found
;	within 1 sec, returns with TOS=0.
;
;	"Key" accepts serial input on EF4
;	with 1 start bit, 8 data, 1 stop,
;	and optionally echoes it via Q.
;
; Pin EF4=0: "1", "stop", or "mark" bit
; Pin EF4=1: "0", "start", or "space" bit
;
; Variables used:
;	baud	- sets baud rate
;	#cycles	- used for 1 sec delay
;	echo=00	- no echo (half duplex)
;	    =01	- echo (full duplex)
;
keys:	glo vbase	;>initialize pointer
	smi 113	  	;   to the variable
	plo ra	  	;   "baud"
	ghi vbase	
	smbi 0
	phi ra
	ldn ra		; get "baud", and	
	str rsp		;   store at TORS	
	inc ra		; point to "#cycles"
	lda ra		; get #cyles
	phi rb		;   and initialize a
	plo rb		;   counter with it
; wait for a stop bit
	ghi rb		; get counter
	dec rb		; decrement counter
	bz timeout	; if 0, go to timeout
	bn4 $-4		; else loop until
			;   a stop bit found
; wait for a start bit
	ghi rb		; get counter
	dec rb		; decrement counter
	bz timeout	; if 0, go to timeout
	b4 $-4		;   else loop until
			;   start bit found
; wait 1/2 bit time
	ldn rsp		; get variable "baud"
	shr		; divide by 2, and:
	str stack	; - save a copy
	smi 1		; - loop for copy*4
	bdf $-2		;   machine cycles
	ldx		; - get copy of baud
	smi 17		; - subtract 17
			;   (for 1/2 bit time)
	bdf $-8		; - repeat until "-"
	b4 $-18		; if start bit gone,
			;   loop and try again
; start bit confirmed
	ldi 10		; set #bits=10 per
	plo rb		;   character
	br echo		; skip ahead to echo

; if no key found within 1 second
timeout: stxd		; push 0 onto stack
	sep next	;<and return to next

; assemble serial bits into a character
nextbit: ldx		; get character from
	shrc		;   TOS, shift in DF
	str stack	;   and save on TOS
	ldn rsp		; get variable "baud"
	phi rb		; - save a copy
	smi 1		; - loop for copy*4
	bdf $-2		;   machine cycles
	ghi rb		; - get copy of baud
	smi 35		; - subract 35
	bdf $-8		; - repeat until "-"
echo:	ldn ra		; get variable "echo"
	bnz $-7		; if echo=0 (no):
	ldn ra	; (nop)	;   then if bit=mark
	b4 $+10		;     then set DF=1
	br $+10		;     else leave DF=0
	b4 $+5		;   else if bit=space
	seq		;     then echo space
	br $+5		;	leave DF=0
	req		;     else echo mark
	smi 0		;	set DF=1
	dec rb		; decrement #bits
	glo rb		; get #bits
	bnz nextbit	; repeat until #bits=0
;
; test for valid stop bit
	bnf $+7		; if stop bit
	ldx		;   get character
	ani 7fh		;   mask to ASCII
	stxd		;   push on to TOS
	sep next	;<  return to next
	stxd		; else push 0=TOS
	sep next	;<return to next

; Q!: Pops TOS and sets Q equal to its
;	Boolean value.
qwri:	inc stack	; pop TOS
	ldx
	bz $+4		; test TOS
	seq		; if = 1, set Q
	sep next
	req		; if = 0, reset Q
	sep next

; DMA!: Pops address from stack and puts
;	it in register R0 (DMA register).
;	TOS=addr.1, NOS=addr.0.
;
dmawri:	irx		;>pop addr.1 and
	ldxa
	phi dmar	;   put in R0.1
	ldx		; pop addr.0 and
	plo dmar	;   put in R0.0
	sep next	;<return to next

; dupnz: Duplicates the top entry on the
;	stack if it is not equal to zero.
;	If TOS=0, "dupnz" has no effect.
;
dupnz:	irx		;>pop TOS
	ldx
	stxd		; push back on stack
	lsz		; if TOS not zero,
	stxd		;   then duplicate
	sep next	;<    return to next
	sep next	;<  else return to next

; append: Expects an address on data stack
;	(TOS=addr.1, NOS=addr.0) & a string
;	(30S=length, 4OS=1st byte, etc.)
;	"Append" removes the 1st byte of the
;	string, & appends it to the end of
;	the string at the specified address.
;
append:	irx		;>initialize pointer
	ldxa		; to address specified
	phi ra		; by TOS, NOS
	ldxa		; - high byte
	plo ra		; - low byte
; test length of string on stack
	ldx		; pop length byte
	lsnz		; if length=0, then
	stxd		;   leave it alone and
	sep next	;<  return to next
	plo rb		; else decr. length
	dec rb		;   and save it
; extract byte from string on stack
	irx		; pop 1st byte of string
	ldx		; from TOS and
	phi rb		;   save it
	glo rb		;   push new length
	stxd		;   on to TOS
; test length of string in memory
	ldn ra		; get mem string length
	adi 1		; and add 1
	lsnf		; if length>255
	sep next	;<  then return to next
	nop		;     (too long)
	str ra		; else store new length
; append byte to string in memory
	str stack	; move pointer to
	glo ra		; end of string:
	add		;   (add length
	plo ra		;   to pointer)
	ghi ra
	adci 0
	phi ra
	ghi rb		; append byte to
	str ra		;   string in memory
	sep next	;<return to next

; D+:	Adds two double-precision numbers (A & B)
;	on the data stack, giving a double-precision
;	(2 byte) sum. The most significant byte is on
;	top of the least significant byte
;	(TOS=A.1, NOS=A.0, etc.
dadd:	irx		;>pop 1st number (A)
	ldxa		; and save it
	phi ra		;   high byte A.1
	ldxa	
	plo ra		;   low byte A.0
	ldxa		; pop 2nd number (B)
	phi rb		;   high byte B.1
	glo ra		; get A.0
	add		; add B.0
	stxd		; push sum.o to TOS
	ghi rb		; get B.1
	str stack
	ghi ra		; get A.1
	adc		; add with carry
	stxd		; push sum.1 to TOS
	sep next	;<return to next

; $! (string write): Expects an address
;	and a string on the stack: TOS=addr.1,
;	NOS=addr.0, 3OS=length of string,
;	4OS=first string character, etc.
;	"$!" stores string in the specified
;	address. The string must be shorter
;	than the number of bytes allotted for
;	that address. No error checking is done.
;
stwri:	irx		;>initialize pointer
	ldxa		; to addr on TOS, NOS
	phi ra	
	ldxa		; - high byte
	plo ra		; - low byte
	ldx		; pop length byte
	plo rb		; & save for later
; move string from stack to memory
	ldxa		; - pop byte of string
	str ra		; - store via pointer
	inc ra		; - increment pointer
	glo rb		; - get length byte
	dec rb		; - decrement length
	bnz $-5		; - repeat until done
	dec stack	; balance stack pointer
	sep next	;<return to next

; $L (string load): Gets the string immediately
;	following via the PPC, & copies it on
;	to the data stack.
; $@ (string at): Expects an address on the stack
;	(TOS=addr.1, NOS=addr.0) pointing to the
;	length byte of a string in memory.
;	"$@" pops the address, and copies the
;	string onto the data stack.
;
stlod:	glo ppc		;>put pseudo-program
	str stack	;   counter.0 on TOS
	ldn ppc		; get string length
	add		; add PPC.0+length
	plo ppc		; put sum in PPC.0
	dec stack	; save PPC.0 on TOS
	ghi ppc		; put pseudo-program
	stxd		; counter.1 on TOS
	adci 0		; add carry to PPC.1
	phi ppc		;   put sum in PPC.1
	inc ppc		; increment PPC

; PPC now points to next word's address,
; and stack contains address of string:
; Continue into stat ($@).
;
stat:	irx		;>initialize pointer
	ldxa		; to addr on TOS, NOS
	phi ra	
	ldx		; - high byte
	plo ra		; - low byte
	ldn ra		; get string length
	plo rb		;   save for later
; move pointer to end of string
	add		; add length+addr.0
	plo ra		; put in pointer.0
	ghi ra		; get addr.1 and
	adci 0		; add carry
	phi ra		; put in pointer.1
; move string to data stack
	ldn ra		; - get next byte
	stxd		; - push onto stack
	dec ra		; - decrement pointer
	glo rb		; - get length byte
	dec rb		; - decrement it
	bnz $-5		; - repeat until done
	sep next	;<return to next

; $drop (string drop): Drops the string
;	on top of the stack. TOS=length,
;	NOS=1st byte of string, etc.
;
stdr:	irx		;>pop length byte
	ldx		;   from TOS
			; repeat...
	irx		; - drop 1 byte
	smi 1		; - length=length-1
	bpz $-3		; - until done
	dec stack	; correct stack pointer
	sep next	;<return to next

; D-: Double-precision subtract of number
;	on TOS (A) from number at NOS (B).  
;	The most significant byte is on top of
;	the least significant byte (TOS=A.1,
;	NOS=A.0, 3OS=B.1, 4OS=B.0).
;
dsub:	irx		;>pop 1st number (A)
	ldxa		; and save it
	phi ra		;   high byte A.1
	ldxa	
	plo ra		;   low byte A.0
	ldxa		; pop 2nd number (B)
	phi rb		;   high byte B.1
	glo ra		; get A.0
	sd		; subtract B.0-A.0
	stxd		; push result.0 to TOS
	ghi rb		; get B.1
	str stack	
	ghi ra		; get A.1
	sdb		; subtract with borrow
	stxd		; push result.1 to TOS
	sep next	;<return to next

; word:	Removes 1st word from the string
;	at address (TOS=addr.1, NOS=addr.0)
;	and moves it to the stack. A word
;	is defined as any string ending
;	with the delimiter specified by
;	3OS. Leading delimiters are dropped.
;	If no delimiter, the entire string
;	is moved to the stack.
;
word:	irx		;>pop address
	ldxa		; set pointer "A" to address
	phi ra		; (length byte of string)
	phi rb
	ldxa		; set pointer "B" to address+1
	plo ra		; (1st byte of string)
	plo ra
	inc rb
	glo ppc		; save low byte of
	str rsp		;   PPC on return stack
	ldi 0		; set count=0
	plo ppc
; locate 1st word and push onto return stack
w1:	ldn ra		; repeat until word
	bz rest		; is found or string
	smi 1		; length=0
	str ra		; - decrement length
	lda rb		; - get next byte
	xor		; - if not delimiter,
	bz delimitr
	inc ppc		;   then increment count
	xor	
	inc rsp		;     and push byte to
	str rsp		;     return stack
	br w1
delimitr: glo ppc	; - if delimiter,
	bz w1		;   ignore if count=0
			;   (leading delimiter)
; word found: move rest of string down
rest:	ldn ra		; let rest=length of rest
	bz moveword	;   of string, and
w2:	smi 1		; repeat until 0:
	str stack	; - decrement rest
	lda rb		; - get byte via RB
	inc ra		; - advance pointer
	str ra		; - store via RA and
	ldx		;   advance pointer
	bnz w2		; - until rest=0

; move word from return stack to stack
;
moveword: glo ppc	; let #bytes=count
	plo ra	
	bz worddone	; repeat...
w3:	ldn rsp		; - pop byte from
	dec rsp		;   return stack
	stxd		; - push onto stack
	dec ra		; - decrement #bytes
	glo ra	
	bnz w3		; - until #bytes=0
worddone: glo ppc	; push count
	stxd		; (=word length) to TOS
	ldn rsp		; restore low byte
	plo ppc		;   of ppc
	sep next	;<return to next

; find:	Searches the vocabulary specified by the
;	address (TOS=addr.1, NOS=addr.0) for the
;	string beginning on the stack at 3OS. If
;	not found, "Find" returns with TOS=false
;	(=0) and the string still on the stack.
;	If found, returns TOS=true (=1), and an
;	address (NOS=addr.1, 3OS=addr.0) pointing
;	to the byte following the end of the name
;	string in the dictionary. The contents of
;	this address (& address+1) is the address
;	of the executable code for the word.
;
;	A vocabulary is a subset of the dictionary,
;	organized as a list of word definitions.
;	Each definition begins with a string giving
;	its name and followed by the adress of the
;	executable code for that word. The entry
;	point into a vocabulary for "find" is at
;	the lowest address (containing the length
;	byte of its newest word). The end of a
;	vocabulary is marked by the null string
;	(=0) followed by address=0000. If a null
;	entry is found, and the number after it
;	is not 0, then "find" will use the number
;	as a link to where the vocabulary is
;	continued. The link is to the address of
;	the length byte of the next entry in the
;	same vocabulary.
;
find:	irx		;>pop address from
	ldxa		; stack and initialize 
	phi ra		; a pointer to
	ldx		; the dictionary
	plo ra		; (=D.ptr)
	sex rb
nextword:		; test words for a match
	ghi stack	; initialize pointer
	phi rb		; to word on stack
	glo stack	; (=W.ptr)
	plo rb
	ldn ra		; get length of word
	bnz nextb	; if=0,
	inc ra		; then let link=
	lda ra		;   next 2 bytes
	bnz getlink	
	ldn ra
	bnz getlink	; if link=0,
	sex stack	;   then word not found:
	stxd		;     push false (=0)
	sep next	;<    return to next
getlink: ldn ra		;   else
	plo rb		;     set D.pointer=link
	dec ra
	ldn ra
	phi ra	
	glo rb
	plo ra		;     and continue
	br nextword-1	;     at next word
;
nextbyte: ; compare dictionary word to stack word
	ldn stack	; let #bytes=length
	bz found	; repeat:
	smi 1		; - decrement #bytes
nextb:	str stack	;   left to compare
	lda ra		; - get byte @ D.ptr.
	irx		; - get byte @ W.ptr.
	xor		; - compare bytes
	bz nextbyte	
;
; if not a match, skip over rest of word
nextn:	inc ra		; repeat:
	ldn stack	; - increment D.ptr.
	smi 1		; - decrement #bytes
	str stack	;   remaining
	bnz nextn	; - until end of word
	inc ra		; increment D.ptr.
	br nextword	; go to next word
;
; word found in vocabulary
found:	sex stack	; drop word on stack
	irx		; let length=#bytes,
	ldx		; and repeat:
	irx		; - drop a byte
	smi 1		; - decrement length
	bpz $-3		; - until length < 0
	dec stack
	glo ra		; push dictionary
	stxd		; pointer to stack
	ghi ra		; (points to addr)
	stxd
	ldi 1		; push true(=1) to
	stxd		; top of stack
	sep next	;<return to next

; dover: Copies 3rd and 4th on stack and
;	pushes them onto TOS and NOS.
;
dover:	ghi stack	;>copy stack
	phi ra		; pointer to RA
	glo stack
	plo ra
	inc ra		; point RA to 4th
	inc ra		; on stack
	inc ra
	inc ra
	ldn ra		; load 4th on stack
	stxd		; push it
	dec ra
	lda ra		; load 3rd on stack
	stxd		; push it
	sep next	;<return to next

; lf:	Tests TOS for ASCII line feed (LF)
;	or carriage return (CR). If LF,
;	replaces TOS with null (0). If CR,
;	output a LF, and leave CR on TOS.
;	All other values are left on TOS
;	unchanged.
;
lf:	irx		;>pop top of stack
	ldx
	smi 10		; test for LF
	lsnz		; if LF,
	stxd		;   then push null
	sep next	;<  return to next
	dec stack	; else push back on TOS
	smi 13-10	;   test for CR
	lsz		; if not CR,
	sep next	;<  return to next
	nop
	ldi 10		; else push LF on TOS
	stxd
	sep call	; output it
	dw emit
	dw return	;<return

; execute: Pops an address from the
;	stack (TOS=addr.1, NOS=addr.0),
;       and executes the program at that
;       address with program counter=R3
;       and X=R2.
;
execute: glo r3		;>set RA=R3
	adi 4
	plo ra		; - low byte
	ghi r3
	phi ra		; - high byte
	inc ra		;   (RA points here)
	sep ra		; set RA=prog.counter
	irx		; pop address from
	ldxa		;   stack & put in R3
	phi r3		; - low byte
	ldx
	plo r3		; - high byte
	sep r3		;<set R3=prog.counter

; * (multiply): Multiplies the top two
;	bytes on the stack (TOS=A, NOS=B)
;	and returns the least significant
;	byte of their product on tos.
;
mult:	irx		;>pop "A" from TOS,
	ldxa
	shr		;   shift right,
	plo ra		;   and put in RA.0
	ldi 0		; set product=0
	phi ra
	ldi 8		; let i=bit counter
	plo rb		; repeat:
	ghi ra		; - get product
	bnf $+3		; - if bit A(8-i)=1,
	add		;   product=product+B
	shrc		; - shift product
	phi ra		;   right
	glo ra		; - get next bit
	shrc		;   of "A"
	plo ra
	dec rb		; - decrement i
	glo rb		; - repeat for all
	bnz $-11	;   8 bits of "A"
	glo ra		; get product, and
	stxd		; push onto TOS
	sep next	;<return to next

; *' (multiply with carry): Multiply the
;	top 2 bytes on the stack (TOS=A,
;	NOS=B), & then adds the carry
;	from a preceeding multiplication
;	(3OS=carry). The double-precision
;	result is returned on the stack
;	(TOS=result.1, NOS=result.0).
;
multc:	irx		;>pop "A" from TOS,
	ldxa
	shr		; shift right,
	plo ra		; and put in RA.0
	ldxa		; pop "B" from nos,
	plo rb		;   and save it
	ldx		; pop carry from 3OS
	phi ra		;   and put in product
	glo rb		; put "B" back on
	str stack	;   stack
	ldi 8		; let i=bit counter
	plo rb		; repeat:
	ghi ra		; - get product
	bnf $+3		; - if bit a(8-i)=1,
	add		;   product=product+B
	shrc		; - shift product
	phi ra		;   right
	glo ra		; - get next bit
	shrc		;    of "A"
	plo ra
	dec rb		; - decrement i
	glo rb		; - repeat for all
	bnz $-11	;   8 bits of "a"
	glo ra		; get product, and
	stxd		; push onto tos
	ghi ra		; - low byte, then
	stxd		; - high byte
	sep next	;<return to next

; $dup (string dup): Duplicate string on
;	stack (TOS=length, NOS=1st byte
;	of string, etc.)
;
stdup:	irx		;>initialize pointer
	glo stack	;   to last byte of
	add		;   string (pointer=
	plo ra		;   stack pointer+
	ghi stack	;   length byte)
	adci 0
	phi ra
	ldx		; let #bytes=length
	plo rb		;   byte of string
	dec stack
			; repeat:
	ldn ra		; - get next byte
	dec ra
	stxd		; - push onto stack
	glo rb		; - get #bytes
	dec rb		;   & decrement
	bnz $-5		; - until #bytes=0
	sep next	;<return to next

; digit: Expects TOS=base and NOS=ASCII
;	character. If the character is a
;	digit in the current base, its
;	binary value is returned at NOS,
;	and TOS=true(=1). If character is
;	not a digit, returns TOS=0(false).
;
digit:	irx		;>pop base from TOS
	ldxa		;   and save it
	plo ra
	ldx		; pop character
	smi '0'		; if < ASCII '0',
	bm notdigit	;   go to notdigit
	smi 10		; if between ASCII
	bm $+6		;   '9' and 'A',
	smi 7
	bm notdigit	;   go to notdigit
	adi 10		; else convert character
	str stack	;   to binary
	glo ra		; get base
	sd		; if binary < base,
	bpz $+6		;   then is a digit;
	dec stack	;   push binary
	ldi 1		;   push true
	lskp
notdigit: ldi 0		; else not a digit
	stxd		;   push false
	sep next	;<return to next

; +' (add with carry): Adds the top 3
;	bytes on the stack, and returns a
;	double-precision sum. It is used
;	to add two bytes (TOS=A, NOS=B) &
;	a carry-in (=3OS) from a previous
;	addition, and return an 8-bit sum
;	(=NOS) and carry-out (=TOS) for
;	extended-precision addition.
;
addc:	ldi 0		;>let carryflag=0
	plo ra
	irx		; pop "A" (from TOS)
	ldxa
	add		; add "B" (from NOS)
	bnf $+3		; if a carry, then
	inc ra		;   set carryflag=1
	irx		; add carry-in
	add		;   (from 3OS)
	stxd		; push sum (to NOS)
	glo ra		; get carryflag,
	adci 0		;   add carry, and
	stxd		;   push carry-out
	sep next	;<return to next

; / (divide): Divides a double-precision
;	dividend (NOS=dividend.1, 3OS=
;	dividend.0) by a single-precision
;	divisor (=TOS), and returns the
;	quotient & remainder on the stack
;	(TOS=quotient, NOS=remainder). If
;	the quotient is larger than 255
;	(hex 'ff') or if divide by 0 is
;	attempted, a quotient and remainder
;	of 255 (hex 'ff') are returned as
;	an error indication.
;
divide: irx		;>pop divisor
	ldxa		;   and save it
	phi rb
	glo stack	; set pointer to
	plo ra		;   dividend.0
	ghi stack	; (leave stack pointer
	phi ra		;   pointing to
	inc ra		;   dividend.1)
	ldi 8		; let i=8 (8 bits)
	plo rb
; assume dividend=dividend/256:
	ghi rb		; if dividend-
	sd		;     divisor>=0
	bm div1		; then quotient>255;
	ldi 255		;   load 255(hex ff)
	str ra		;   into quotient
	stxd		;   & remainder, &
	sep next	;<  return to next
; divide via shift and subtract
div3:			; repeat...
	dec rb		; - decrement i
	ldx		; - multiply dividend
	shlc		;   by 2 (shift left)
	str stack
	ghi rb
	bnf div2	; - if dividend>255,
	sd		;   then subtract the
	str stack	;     divisor, & set
	smi 0		;     quotient(i)=1
	br div1		;   else
div2:	sd		;     if dividend-
	bm div1		;	  divisor>=0,
	str stack	;     then subtract
div1:	ldn ra		;	divisor & set
	shlc		;	quotient(i)=1
	str ra		;     else
	glo rb		;	quotient(i)=0
	bnz div3	; - until i=0
	dec stack	; push remainder
	sep next	;<return to next

; d@ (double precision "@"): Expects 2 bytes
;	on data stack, specifying an address
;	(TOS=addr.1, NOS=addr.0). d@ reads the
;	2 bytes at addr and addr+1, and pushes
;	them onto TOS and NOS respectively.
;
dat:	irx		;>pop address from
	ldxa		;   stack
	phi ra
	ldx
	plo ra
	inc ra		; read address+1 and
	ldn ra		;   push contents
	stxd		;   onto stack
	dec ra		; read address and
	ldn ra		; push contents
	stxd		;   onto stack
	sep next	;<return to next

; d! (double precision "!"): Expects 4
;	bytes on data stack; TOS=addr.1,
;	NOS=addr.0, 3OS=data.1, and 4OS=
;	data.0. "d!" writes the 2 data
;	bytes into memory starting at
;	the specified address.
;
dwri:	irx		;>pop address from
	ldxa		;   stack
	phi ra
	ldxa
	plo ra
	ldxa		; pop data.1 and
	str ra		;   store at address
	inc ra
	ldx		; pop data.0 and
	str ra		;   store at addr+1
	sep next	;<return to next

; sp@:	Sets the address of where the
;	current stack pointer points and
;	pushes it onto the stack (tos=
;	addr.1, nos=addr.0).
;
spat:	ghi stack	;>save a copy of
	phi ra		;   stack pointer.1
	glo stack	;   in RA.1
	stxd		; push stack pointer.0
	ghi ra		;   push saved copy
	stxd		;   from RA.1
	sep next	;<return to next

; rsp@:	Gets the address pointed to by
;	the current return stack pointer
;	and pushes it on to the stack
;	(NOS=addr.1, TOS=addr.0).
;
rspat:	glo rsp		;>push addr.1 onto
	stxd		;   data stack
	ghi rsp		; push addr.0 onto
	stxd		;   data stack
	sep next	;<return to next

; d>r:	Pops a double-precision number
;	(TOS=number.1, NOS=number.0),
;	then pops the specified number
;	of bytes from the data stack and
;	pushes them on the return stack.
;
datamove: irx		;>pop #bytes to be
	ldxa		;   moved (double-
	phi ra		;   precision)
	ldx
	plo ra
dmove1: glo ra		; repeat...
	bnz $+5
	ghi ra
	bz movedone
	irx		; - pop TOS
	ldx
	str rsp		; - push onto TORS
	inc rsp
	dec ra		; - decrement #bytes
	br dmove1	; - until #bytes=0
movedone: sep next	;<return to next

; d<:	Pops 2 double-precision numbers
;	(A and B) from the stack and
;	compares them: TOS=A.1, NOS=A.0,
;	3OS=B.1, 4OS=B.0. If B<A, returns
;	with TOS = true(=1), else with
;	TOS = false(=0).
;
dless:	irx		;>pop A and save in RA
	ldxa
	phi ra		; - high byte
	ldxa
	plo ra		; - low byte
	ldxa		; pop B.1 and save
	phi rb		;   in RB
	glo ra		; subtract: B.0-A.0
	sd
	ghi rb		; get B.1
	str stack
	ghi ra		; get A.1
; subtract B.1-A.1-(borrow)
; if borrow is true (DF=0), then B<A
	sdb
	shlc		; shift borrow into D
	xri 1		; invert D
	ani 1		; mask rest of D to 0
	stxd		; TOS gets result
	sep next	;<return to next

; dswap (double-precision swap): Swaps
;	double-precision number at TOS &
;	NOS (TOS=num.1, NOS=num.0) with
;	double precision number at 3OS &
;	4OS (3OS=num.1, 4OS=num.0).
;
dswap:	irx		;>pop number at TOS
	ldxa		;   & NOS into RA
	phi ra
	ldxa
	plo ra
	ldxa		; pop number at 3OS
	phi rb		;   & 4OS into RB
	ldx
	plo rb
	glo ra		; push number that is
	stxd		;   in RA at 3OS
	ghi ra		;   and 4OS
	stxd
	glo rb		; put number that is
	stxd		;   in RB at TOS
	ghi rb		;   and NOS
	stxd
	sep next	;<return to next

; r>d:	Pops a double-precision number
;	(NOS=number.0, TOS=number.1),
;	then pops the specified number
;	of bytes from the return stack &
;	pushes them onto the data stack.
;
rtnmove: irx		;>pop #bytes to be
	ldxa		;   moved (double-
	phi ra		;   precision)
	ldx
	plo ra
rmove1: glo ra		; repeat...
	bnz $+5
	ghi ra
	bz rmovdone
	dec rsp		; - pop TORS
	ldn rsp
	stxd		; - push onto TOS
	dec ra		; - decrement #bytes
	br rmove1	; - until #bytes=0
rmovdone: sep next	;<return to next

; maxram: Pushes the address of the
;	variable "maxram" (TOS=addr.1,
;	NOS=addr.0).
;
maxram: sep variable	; push address
	dw 1		;   of "maxram"

; minram: Pushes the address of the
;	variable "minram" (TOS=addr.1,
;	NOS=addr.0).
;
minram: sep variable	; push address
	dw 3		;   of "minram"

; sp0:	Pushes the address of the
;	variable "sp0" (TOS=addr.1,
;	NOS=addr.0).
;
sp0:	sep variable	; push address
	dw 5		;   of "sp0"

; rsp0: Pushes the address of the
;	variable "rsp0" (TOS=addr.1,
;	NOS=addr.0).
;
rsp0:	sep  variable	; push address
	dw 7		;   of "rsp0"

; first: Pushes the address of the
;	variable "first" (TOS=addr.1,
;	NOS=addr.0).
;
first:	sep variable	; push address
	dw 9		;   of "first"

; next: Pushes the address of the
;	variable "next" (TOS=addr.1,
;	NOS=addr.0).
;
next2:	sep variable	; push address
	dw 11		;   of "next"

; call:	Pushes the address of the
;	variable "call" (TOS=addr.1,
;	NOS=addr.0).
;
call2:	sep variable	; push address
	dw 13		;   of "call"

; interrupt: Pushes the address of the
;	variable "interrupt" (TOS=addr.1,
;	NOS=addr.0).
;
intrupt1: sep variable	; push address
	dw 15		;   of "interrupt"

;
; vat:	Pushes address of the variable
;      "v@" (TOS=addr.1, NOS=addr.0).
;
vat:	sep variable	; push address
	dw 17		;   of "vat"

; dma:	Pushes the address of the
;	variable "dma" (TOS=addr.1,
;	NOS=addr.0).
;
dma2:	sep variable	; push address
	dw 19		;   of "dma"

; current: Pushes the address of the
;	variable "current" (TOS=addr.1,
;	NOS=addr.0).
;
current: sep variable	; push address
	dw 21		;   of "current"

; context: Pushes the address of the
;	variable "context" (TOS=addr.1,
;	NOS=addr.0).
;
context: sep variable	; push address
	dw 23		;   of "context"

; n:	Pushes the address of the variable
;	"n" (TOS=addr.1, NOS=addr.0).
;
n:	sep variable	; push address
	dw 25		;   of "n"

; base:	Pushes address of the variable
;	"base" (TOS=addr.1, NOS=addr.0).
;
base:	sep variable	; push address
	dw 26		;   of "base"

; Terminator: Pushes the address of the
;	variable "terminator" (TOS=addr.1
;	NOS=addr.0).
;
termnatr: sep variable	; push address
	dw 27		;   of "terminator"

; delimiter: Pushes the address of the
;	variable "delimiter" (TOS=addr.1,
;	NOS=addr.0).
;
delmiter: sep variable	; push address
	dw 28		;   of "delimiter"

; buffer: Pushes the address of the
;	variable "buffer" (TOS=addr.1,
;	NOS=addr.0).
;
buffer:	sep variable	; push address
	dw 109		;   of "buffer"

; caps: Pushes the address of the
;	variable "caps" (TOS=addr.1,
;	NOS=addr.0).
;
caps:	sep variable	; push address
	dw 110		;   of "status"

; echo: Pushes the address of the
;	variable "echo" (TOS=addr.1,
;	NOS=addr.0).
;
echo1:	sep variable	; push address
	dw 111		;   of "echo"

; #cycles: Pushes the address of the
;	variable "#cycles" (TOS=addr.1,
;	NOS=addr.0).
;
nocycles: sep variable	; push address
	dw 112		;   of "#cycles"

; baud: Pushes address of the variable
;	"baud" (TOS=addr.1, NOS=addr.0).
;
baud:	sep variable	; push address
	dw 113		;   of "baud"

; lastkey$: Pushes the address of the
;	variable "lastkey$" (TOS=addr.1,
;	NOS=addr.0).
;
lastkeyn: sep variable	; push address
	dw 114		;   of "lastkey$"

; $abort: Expects a string on the stack.
;	Pops the string, places it in the
;	buffer, and aborts.
;
stabort: sep call	; call 8TH
	dw buffer	; write string on
	dw stwri	;   stack into buffer
	dw abort	; abort (prints string)
	dw return	; return to next

; base@: Pushes the value of "base"
;	onto TOS.
;
basat:	sep call	; call 8TH
	dw base		; get "base"
	dw at
	dw return	; return to next

; curnt@: Pushes the address of the
;	variable that "current" points to.
;
curntat: sep call	; call 8TH
	dw current	; get "current"
	dw dat
	dw return	; return to next

; input: Pushes the address of the
;	variable "input" (TOS=addr.1,
;	NOS=addr.0).
;
input:	sep variable	; call variable
	dw 116

; output: Pushes the address of the
;	variable "output" (TOS=addr.1
;	NOS=addr.0).
;
output:	sep variable	; call variable
	dw 118

; emit: Outputs TOS as an ASCII character
;	to the output routine specified
;	by the variable "output".
;
emit:	sep call	; call 8TH
	dw output	; get output routine
	dw dat		; address
	dw execute	; execute it
	dw return	; return to next

; type: Types out the string on the stack
;	to the output device specified
;	by the variable "output".
;
type:	sep call	; call 8TH
	dw dupnz	; do for length of
	dw br0		;   string:
	dw typ1
	dw 1
	dw do
	dw emit		; - output a char
	dw loop
typ1:	dw return	; return to next

; cr:	Types out a carriage return (CR)
;	and line feed (LF) to the output
;	device.
;
cr:	sep call	; call 8TH
	dw stlod	; push the string
	db 2,13,10	;   ...ASCII CRLF
	dw type		; type it
	dw return	; return to next

; memful: Checks for memory being full.
;	If it is, returns with TOS=1,
;	else TOS=0. If memory if full,
;	prints error message.
;
memful:	sep call	; call 8TH
	dw spat		; get TOS address
	dw rspat	; get TORS address
	dw 32		; reserve 32 bytes
	dw 0		;   as a pad
	dw dadd
	dw dless	; set TOS
	dw dup
	dw br0		; if memory full
	dw memf1	;   type error mesg
	dw stlod
	db 12,' MEMORY FULL'
	dw type
	dw cr
memf1:	dw return

; check: Checks the data and return
;	stacks for errors, if any are
;	found, puts error message in the
;	buffer, and aborts. Error
;	conditions are data stack empty,
;	return stack empty, and memory
;	full (stacks overlap).
;
check:	sep call	; call 8TH
	dw sp0		; get base address
	dw dat		;   of data stack
	dw spat		; get TOS address
	dw 8		; reserve 8 bytes as
	dw 0		;   pad between stack
	dw dadd		;   and dictionary
	dw dless	; if TOS > SP0,
	dw br0
	dw check1	;   push message...
	dw stlod
	db 12,' STACK EMPTY'
	dw stabort	;   and abort
check1:	dw rspat	; get TORS address
	dw rsp0		; get address of base
	dw dat		;   of return stack
	dw dless	; if TORS <RSP0,
	dw br0
	dw check2
	dw stlod	;   push message...
	db 19,' RETURN STACK EMPTY'
	dw stabort	;   and abort
check2:	dw memful	; test memory space
	dw br0
	dw check3
	dw abort	;   abort if full
check3	dw return	; else return to next

; getakey: Gets a key from the keyboard
;	input device, and returns it on
;	TOS. If carriage return, a line
;	feed automatically follows. Nulls
;	and line feeds are ignored.
;
getakey: sep call	; call 8TH
	dw repeat	; repeat...
	dw input	; - input a key from
	dw dat		;     device specified
	dw execute	;     by "input"
	dw lf		; - if CR, add LF
	dw dupnz
	dw until	; - until done
	dw return

; expect: Gets keys from keyboard until
;	either the number of keys specified
;	by TOS, or a carriage return or
;	cancel key is found. The keys are
;	placed in the string variable
;	"buffer". "Key" returns TOS=true
;	if the specified number of keys
;	was found, or TOS=false if a
;	carriage return or cancel ended
;	the input. The carriage return or
;	cancel is not returned as part of
;	the string.
;
expect:	sep call	; call 8TH
	dw 0		; clear the buffer
	dw buffer
	dw excla
	dw dup		; if the number of
	dw br0		;   keys requested
	dw e1		;   is not zero,
	dw repeat	; then repeat...
	dw getakey	;   - get a key
	dw dup
	dw backspce	;   - if backspace,
	dw at
	dw equal
	dw br0
	dw e2
	dw buffer	;     then remove
	dw at		;     the last key
	dw dupnz	;     from the
	dw br0		;     buffer
	dw e7
	dw 1
	dw subtr
	dw buffer
	dw excla
e7:	dw br
	dw e3
e2:	dw dup		;   - if cancel,
	dw cancel
	dw at
	dw equal
	dw br0
	dw e4
	dw stlod	;	type "<-CANCEL"
	db 9,' <-CANCEL'
	dw type
	dw cr
	dw 0		;	and clear the
	dw buffer	;	buffer
	dw excla
	dw br
	dw e3
e4:	dw dup		;   - if terminator,
	dw termnatr
	dw at
	dw equal
	dw br0
	dw e5
	dw drop		;	then put a zero
	dw 0		;	on TOS to exit
	dw swap		;	at next chance
	dw br
	dw e3
e5:	dw 1		;   - else is data key
	dw buffer	;	so append it to
	dw append	;	the buffer
e3:	dw drop		;   - key is now fully
	dw dup		;     processed: drop
	dw buffer	;     it & get buffer
	dw at		;     length
	dw greater	;   - if requested no.
	dw not1		;     of keys not yet
	dw until	;     found, repeat
	dw br
	dw e6
e1:	dw drop		; else no keys requested:
	dw 1		;   set TOS=true
e6:	dw return	;   and return

; dnumber: Expects a string on the stack.
;	stack. If it is a number in the
;	current base, it is converted to
;	binary and returned as a double-
;	precision number (TOS=true, NOS=
;	number.1, 3OS=number.0. If it
;	can't be converted, the string is
;	returned on the stack, and TOS=
;	false.
;
dnumber: sep call	; call 8TH
	dw stdup	; duplicate string
	dw dup		; if not null
	dw br0		;   string, then:
	dw dnum1
	dw 0		;   set n=0 (double-
	dw 0		;     precision
	dw n
	dw dwri
	dw dup		; do for i=1 to end
	dw 1		; of string...
	dw do
	dw 1		; - remove next
	dw subtr	;   character from
	dw swap		;   string
	dw basat	; - get base
	dw digit	; - if character is
	dw br0		;   a digit in the
	dw dnum2	;   current base,
	dw n		;   then calculate
	dw dat		;   n=(n*base)
	dw drop		;   +digit
	dw basat
	dw multc
	dw n
	dw at
	dw basat
	dw multc
	dw br0		; - if n overflows,
	dw dnum3
	dw ddrop	;   then drop n,
	dw stdr		;   drop string,
	dw 0		;   push false,
	dw exit		;   and exit
	dw br
	dw dnum4
dnum3:	dw n		; - else store new
	dw dwri		;   value of n
	dw dupnz	; - if at end of
	dw not1		;   string, is
	dw br0		;   successful
	dw dnum4	;   conversion
	dw stdr		; - drop the string
	dw n
	dw dat		; - push n,
	dw 1		; - & push true
dnum4:	dw br
	dw dnum5	; - else not a digit,
			;     so string is not
			;     a number:
dnum2:	dw stdr		; - drop string,
	dw 0		; - push false,
	dw exit		; - and exit
dnum5	dw loop		; - loop until done
dnum1	dw return	; return

; number: Expects a string on the stack.
;	If it is a number in the current
;	base, it is converted into binary
;	& returned as a single-precision
;	number (TOS=true, NOS=number).
;	If it can't be converted into a
;	1-byte number, the string is
;	returned on the stack, & TOS=false.
;
number:	sep call	; call 8TH
	dw dnumber	; convert to double
	dw br0		;   precision number
	dw num1		; if it can be
	dw dupnz	;   converted
	dw br0		;   but not single
	dw num2		;   precision, return
	dw tonumber	;   with string
	dw 0		; push false
	dw br		; if single precision,
	dw num3
num2:	dw 1		;   push true, leave #
num3:	dw br		;   alone
	dw num4		; else push 0 (cannot
num1:	dw 0		:   convert to a #)
num4:	dw return	; return to next

; ddup (double precision "dup"): Duplicates
;	the top two bytes on the data stack.
;
ddup:	sep call	; call 8TH
	dw over		; copy NOS
	dw over		; copy TOS
	dw return	; return to next

; #:	Converts a double-precision number
;	on the stack (TOS=number.1, NOS=
;	number.0) into an ASCII string in
;	the current base. The string is
;	returned on the stack (TOS=length
;	NOS=1st character, etc.) Leading
;	0's are suppressed. If base<2 the
;	number can't be converted, so the
;	null string (0) is returned.
;
tonumber: sep call	; call 8TH
	dw basat	; get base
	dw 2		; get 2
	dw less		; compare them
	dw br0		; if base<2,
	dw tonum2
	dw ddrop	;   then drop number &
	dw 0		;   return null (0)
	dw br
	dw tonum1	; else,
tonum2:	dw 16		;   for digits 16...
	dw 1		;   ... to 1 ...
	dw do		;   do the following:
	dw 0		;   - divide the number
	dw basat	;     by the base;
	dw divide	;     - divide the high
	dw rot		;       byte,
	dw swap
	dw basat	;     - get base,
	dw divide	;     - divide low byte
	dw dup
	dw 9		;   - convert remainder
	dw greater	;     into ASCII
	dw br0
	dw tonum3	;   - if >9,
	dw 7		;       then add 7
	dw add
tonum3:	dw 48		;   - add 48
	dw add
	dw swap		;   - rotate ASCII
	dw rot		;	under number
	dw ddup
	dw or1		;   - if number=0 (con-
	dw not1		;     version complete)
	dw br0		;     then 
	dw tonum4	;       drop the number
	dw ddrop	;       push length of
	dw i		;	  ASCII string
	dw exit		;	and exit loop
tonum4:	dw loop		;   - else loop until
tonum1:	dw return	;       16 characters

; getword: Removes the next word from the
;	input buffer and pushes it onto the
;	data stack as a string. The variable
;	"delimiter" is used to specify the
;	word's delimiter. If a null is found,
;	indicating end-of-line, getword inputs
;	a new line of text to the buffer and
;	gets the first word.
;
getword: sep call	; call 8TH
	dw repeat
	dw delmiter	; push delimiter to TOS
	dw at
	dw buffer	; push buffer address
	dw word		; get next word
	dw dup
	dw dupnz
	dw not1		; if word=null,
	dw br0		;   get new line of text
	dw get1
	dw repeat
	dw 80
	dw expect
	dw dup
	dw br0		; if buffer full,
	dw get2		;   type message
	dw stlod
	db 14,' LINE TOO LONG'
	dw type
get2:	dw not1
	dw until
get1:	dw until
	dw return	; return to next

; word?: Expects a string on the stack.
;	types the string followed by a
;	"?", then clears the buffer.
;
wordq:	sep call	; call 8TH
	dw type		; type the string
	dw stlod	; type " ?",
	dc 2, ' ?'
	dw type		;   followed by CR
	dw cr
	dw 0
	dw buffer	; clear the buffer
	dw excla
	dw return	; return to next

; number?: Expects a string on the stack.
;	Tries to convert the string into
;	a single-precision number. If it
;	is not able to do the conversion,
;	then the string (with " ?" appended
;	to it) is typed, and TOS is set to 0.
;	If successful, the number is left
;	on the stack and TOS is set to 1.
;
numberq: sep call	; call 8TH
	dw number	; try converting
			;   string to number
	dw br0		; if able to,
	dw numc1	;   set TOS=true
	dw 1
	dw br		; else type word
	dw numc2	;   followed by " ?"
numc1:	dw wordq	;   and set TOS = false
	dw 0
numc2:	dw return

; ' (tick): Expects a string on the stack
;	and searches the current and context
;	vocabulary for it. If found, returns
;	with the addr of the executable code
;	for the word at 3OS & NOS, and TOS=1.
;	If not found, returns with the string
;	starting at NOS, and TOS=0
;
tick:	sep call	; call 8TH
	dw getword	; get next word, and
	dw curntat	;   search current
	dw dat		;   vocabulary
	dw find
	dw dupnz	; if not found,
	dw not1		;   search context
	dw br0		;   vocabulary
	dw ti1
	dw context
	dw dat
	dw dat
	dw find
ti1:	dw br0		; if word was found
	dw ti2		;   in either vocab,
	dw dat		; then return its
	dw 1		;   executable address
	dw br		;   at NOS and 3OS,
	dw ti3		;   and TOS=1
ti2:	dw 0		; else return w. string
ti3:	dw return	;   at NOS, and TOS=0

; intrpret (outer interpreter): Gets a
;	line of words from the keyboard
;	and executes them.
;
intrpret: sep call	; call 8TH
	dw 0		; reserve 6 bytes
	dw dup		;   at beginning
	dw ddup		;   of stack
	dw ddup
	dw repeat	; repeat...
	dw buffer	; - type message
	dw stat		;   in the buffer
	dw type
	dw cr		; - type CR and LF
	dw 0		; - clear the buffer
	dw buffer
	dw excla
int2:	dw repeat	; - repeat...
	dw tick		;   - get next word,
			;     seek it in the
			;     current and
	dw br0		;     context vocabs.
	dw int4		;   - if found,
	dw execute	;	then try to
	dw br		;	  execute it
	dw int5		;	else
int4:	dw numberq	;	  try to
	dw not1		;	  convert to
	dw br0		;	  a number
	dw int5
	dw abort
int5:	dw check	;   - check stacks
	dw buffer	;   - until buffer
	dw at		;     is empty
	dw not1
	dw until
	dw stlod	; - move "OK" to
	dc 3,' OK'	;   the buffer
	dw buffer
	dw stwri
	dw 0		; repeat forever
	dw until
	dw return

; d. (double precision print): Pops TOS
;	and NOS, and types them as a
;	double precision number.
;
ddot:	sep call	; call 8TH
	dw tonumber	; convert to a string
	dw type		; type it
	dw 32		; type a space
	dw emit
	dw return

; . (print): Pops TOS and types it as a
;	number in the current base
;
dot:	sep call	; call 8TH
	dw 0		; convert to double precision
	dw ddot		; print number
	dw return

; decimal: Sets base to decimal.
;
decimal: sep call	; call 8TH
	dw 10		; set base to 10
	dw base
	dw excla
	dw return	; return to next

; hex:	Sets base to hexadecimal.
;
hex:	sep call	; call 8TH
	dw 16		; set base to 16
	dw base
	dw excla
	dw return	; return to next

; can:	Pushes the address of the variable
;	"cancel" (TOS=addr.1, NOS=addr.0).
;
cancel:	sep variable	; push address
	dw 119		; of "cancel"

; bs:	Pushes the address of the variable
;	"backspace" (TOS=addr.1, NOS=addr.0).
;
backspce: sep variable	; push address
	dw 120		; of "backspace"

; " (quote): Extracts characters from the
;	beginning of the string in "buffer"
;	until a " is found. The string
;	extracted is then pushed onto the
;	stack.
;
quote:	sep call	; call 8TH
	dw 22h		; get word delimited
	dw buffer	;   by ASCII "
	dw word		;   from the buffer
	dw return	; return to next

; size:	Pushes the address of the variable
;	"size" (TOS=addr.1, NOS=addr.0).
;
size:	sep variable	; push address
	dw 122		; of "size"

; user:	Pushes the address of the variable
;	"user" (TOS=addr.1, NOS=addr.0).
;
user:	sep variable	; push address
	dw 124		; of "user"

; 8th:	Pushes the address of the variable
;	"8th" (TOS=addr.1, NOS=addr.0).
;
eighth:	sep variable	; push address
	dw 126		; of "8th"

; compiler: Pushes the address of
;	the variable "compiler"
;	(TOS=addr.1, NOS=addr.0).
;
compiler: sep variable	; push address
	dw 128		; of "compiler"

; datasize: Calculates the current size
;	of the data stack and leaves the
;	result as a double precision
;	number on the stack.
;
datasize: sep call	; call 8TH
	dw spat		; calc difference
	dw sp0		; between base and
	dw dat		; top of data stack
	dw dswap
	dw dsub
	dw return

; d!+ (double precision increment):
;	Increments the double precision
;	number on the stack by 1.
;
dinc:	sep call	; call 8TH
	dw 1,0		; double-precision 1
	dw dadd		; add to stack
	dw return

; erase: Expects a string on the stack.
;	Searches the current vocabulary
;	for the word which matches the
;	string. If found, erases it and
;	all words defined after it.
;
erase:	sep call	; call 8TH
	dw curntat	; search current
	dw dat		;   vocab for word
	dw find		; if found,
	dw br0		;   move bases of
	dw era1		;   each stack back
	dw ddup		;   so the word's
	dw dat		;   dictionary entry
	dw rsp0		;   becomes part of
	dw dwri		;   the data stack,
	dw dinc		;   & its executable
	dw ddup		;   code becomes part
	dw sp0		;   of the return stack
	dw dwri
	dw dinc		;   correct current
	dw curntat	;   vocabulary pointer
	dw dwri
	dw abort	;   abort to clear
			;   stacks & erase word
era1:	dw wordq	; else word not found
	dw abort	;   type it, "?" and CRLF

; forget: Forgets the next word entered
;	into the buffer and all words
;	defined after it.
;
forget:	sep call	; call 8TH
	dw getword	; get next word
	dw stlod
	db 5,' OK',13,10; type " OK" and CRLF
	dw buffer
	dw stwri	; erase word, and
	dw erase	;   exit (via abort)

; create: Gets the next word entered into
;	the buffer and creates a dictionary
;	entry for it that points to the
;	base of the return stack.
;
create:	sep call	; call 8TH
	dw datasize	; get size of data
	dw ddup		;   stack
	dw size		;   and save it
	dw dwri		; 
	dw datamove	; move contents of
	dw rsp0		;   data stack to
	dw dat		;   return stack 
	dw getword	; create dictionary
	dw spat		;   entry (underneath
	dw ddup		;   data stack)
	dw sp0		; set new base of
	dw dwri		;   data stack
	dw dinc
	dw curntat	; update current
	dw dwri		;   vocab pointer
	dw size
	dw dat
	dw rtnmove	; restore data stack
	dw return

; compile: Expects a number on TOS, pops
;	number from TOS, and pops that
;	many bytes from the data stack,
;	and places them under the base
;	of the return stack. This is
;	where the executable code for a
;	new word definition will go.
;
compile: sep call	; call 8TH
	dw dupnz	; pop # at TOS
	dw br0		; do the following
	dw comt1	; # times
	dw 1
	dw do
	dw rspat	; - place a "0"
	dw rsp0		;   under the
	dw dat		;   return stack
	dw dsub
	dw ddup
	dw size
	dw dwri
	dw rtnmove
	dw 0
	dw size
	dw dat
	dw dinc	
	dw datamove
	dw rsp0		; - write TOS to
	dw dat		;   location of "0"
	dw excla
	dw rsp0		; - move RSP0 over
	dw dat		;   new byte (under
	dw dinc		;   return stack)
	dw rsp0
	dw dwri
	dw loop
comt1:	dw return	; return to next

; , (comma): Pops TOS and places it
;	under the return stack. Used for
;	assembling machine language
;	definitions, one byte at a time.
;
comma:	sep call	; call 8TH
	dw 1		; compile 1 byte
	dw compile
	dw return	; return

; allot: Expects a 1 byte size at TOS.
;	Creates a hole at the top of the
;	dictionary equal to the size
;	popped from TOS. Also creates
;	a link around the hole and a
;	routine	that upon execution gives
;	the address of the hole. It does
;	not create a dictionary entry for
;	the routine.
;
allot:	sep call	; call 8TH
	dw dupnz	; get size of hole
	dw br0
	dw all1
	dw datasize	; get size of data
	dw ddup		; stack, save it,
	dw size		;   and move stack
	dw dwri		;   to return stack
	dw datamove
	dw rspat	; retrieve size of
	dw size		; hole from middle
	dw dat		; return stack
	dw dsub	
	dw at
	dw 1		; push the hole
	dw do
	dw 0
	dw loop
	dw sp0		; create a link
	dw dat		; around it
	dw dinc
	dw 0
	dw spat		; set sp0 over hole
	dw sp0
	dw dwri	
	dw size		; get size of
	dw dat		;   data stack
	dw sp0		; point current
	dw dat		;   vocab to the link
	dw dinc
	dw ddup
	dw curntat
	dw dwri
	dw maxram	; compile routine
	dw dat		; that gives holes
	dw dswap	; address
	dw 3,0
	dw dadd
	dw dsub
	dw 0d8h
	dw 3
	dw compile
	dw rtnmove	; restore data stack
	dw drop
all1:	dw return	; return to next

; variable: Expects a 1 byte size at TOS,
;	pops TOS and creates a variable
;	with that size. The name given
;	to the variable is the next
;	word entered in the buffer.
;
varible: sep call	; call 8TH
	dw dupnz	; if size=0, then
	dw br0		;   do not create
	dw vari1	;   a variable
	dw create	; create dict. entry
	dw allot	; allot the space
vari1:	dw return	; return to next

; ; (semi-colon): ";" is a compiler word
;	that marks the end of a ":"
;	definition of a new word. ";"
;	extends the current vocabulary
;	to include the new word, moves
;	RSP0 to save the executable code,
;	and changes TOS=true to exit the
;	compile mode.
;
semicoln: sep call	; call 8TH
	dw dup
	dw 1		; test that all do,
	dw equal	; repeat, and if-
	dw br0		; else structures
	dw semi1	; are complete
	dw drop
	dw A.0(return)	; end new def
	dw A.1(return)	;   with return
	dw 2
	dw compile
	dw 0		; TOS gets ";" code
	dw return	; return to next
semi1:	dw stlod	; if not,
	db 1,';'	;   then print error
	dw wordq
	dw return	; return to next

; " (compiler usable quote): Used in the
;	compile mode to push the string
;	immediately following, delimited
;	by ", onto the data stack.
;	Compiles into a $L followed by
;	the string.
;
quote1:	sep call	; call 8TH
	dw A.0(stlod)	; push address of
	dw A.1(stlod)	;   $L on stack
	dw 2		; compile it under
	dw compile	;   return stack
	dw quote	; get string from
	dw dup		;   from buffer
	dw 1		;
	dw add		; compile string
	dw compile	;   under return
	dw return	;   stack

; if (compiler word): "If" compiles a BR0
;	followed by a 2-byte "hole" into
;	which "else" or "endif" will place
;	the branch address. The address
;	of the lower byte of this hole is
;	pushed onto the data stack.
;
if:	sep call	; call 8TH
	dw 0,0		; compile a "br0"
	dw A.0(br0)	;   followed by a
	dw A.1(br0)	;   2-byte hole
	dw 4
	dw compile
	dw rsp0		; place address of
	dw dat		;   hole on stack
	dw 2,0
	dw dsub
	dw 2		; push "if" code=2
	dw return	; return to next

; else (compiler word): "Else" pops an
;	address from TOS and NOS, and
;	places the address of the
;	location immediately following
;	this compiled "else" (and its
;	branch address) in the address
;	that was popped. This popped
;	address points to the location
;	of a branch address of a
;	corresponding "if".
;
else:	sep call	; call 8th
	dw dup
	dw 2		; test for "if" code=2
	dw equal	;   to be sure an "if"
	dw br0		;   preceeded "else"
	dw else1
	dw drop		; if true,
	dw rsp0
	dw dat
	dw 4,0		;   write addr into
	dw dadd		;     BR0(addr)
	dw dswap	;     compiled 
	dw dwri		;     by "if"
	dw 0,0		;   compile a
	dw A.0(br)	;     "BR(addr)"
	dw A.1(br)	;     with 2-byte
	dw 4		;     hole for addr
	dw compile
;pg 62
	dw rsp0		;   place addr of
	dw dat		;     hole on stack
	dw 2,0
	dw dsub
	dw 3		;   push "else" code=3
	dw return	;   return
else1:	dw stlod	; if false,
	db 4,'ELSE'	;   print error msg
	dw wordq
	dw return	;   return

; endif (compiler word): "Endif" closes
;	an "if-else-endif" structure.
;	Pops an address from TOS and NOS
;	and places the address of the
;	location where the endif is to
;	compile at the address that was
;	popped. This address points to a
;	branch address for a corresponding
;	"if" or "else". No code is
;	actually compiled for an endif.
;
endif:	sep call	; call 8th
	dw dup
	dw dup		; test for "if" code=2
	dw 2		;   to be sure "if" or
	dw equal	;   "else" preceeded
	dw not1		;   this "endif"
	dw br0
	dw endf1
	dw 3
	dw equal
	dw not1
	dw br0		; if false,
	dw endf2	;   print error
	dw stlod	;   message
	db 5,'ENDIF'
	dw wordq
	dw return	;   return
endf2:	dw drop
	dw br
	dw endf3
endf1:	dw ddrop	; if true,
endf3:	dw rsp0		;   write BR addr
	dw dat		;   to "hole" addr
	dw dswap	;   on stack
	dw dwri
	dw return	;   return

; do (compiler word): "Do" compiles the
;	executable code for "do" under
;	the base of the return stack.
;	Sets TOS=4 as an indication that
;	this compiler word occurred.
;
do1:	sep call	; call 8th
	dw A.0(do)	; compile addr of
	dw A.1(do)	;   "do" under rtn
	dw 2		;   stack
	dw compile
	dw 4		; push "do" code=4
	dw return	; return to next

; loop (compiler word): "Loop" compiles
;	the executable code for the word
;	"loop" under the return stack.
;	Before compilation, tests to make
;	sure that TOS=4, which indicates
;	that there exists a corresponding
;	"do".
;
loop1:	sep call	; call 8th
	dw dup
	dw 4		; test for "do" code=4
	dw equal	;   to be sure a "do"
	dw br0		;   preceeded "loop"
	dw loop11	; if true,
	dw drop
	dw A.0(loop)	;    compile "loop"
	dw A.1(loop)	;    under rtn stk
	dw 2	
	dw compile	
	dw return	;    return
loop11:	dw stlod	; if false,
	db 4,'LOOP'	;    print error msg
	dw wordq
	dw return	;    return

; +loop (compiler word): "+Loop" compiles
;	the address of the executable code
;	for the word "+loop" under the
;	return stack. Before compilation,
;	tests to make sure that TOS=4,
;	which indicates that there exists
;	a corresponding "do".
;
pluslp1: sep call	; call 8th
	dw dup
	dw 4		; test for "do" code=4
	dw equal	;   to be sure a "do"
	dw br0		;   preceeded "+loop"
	dw plu1		; if true,
	dw drop
	dw A.0(pluslp)	;   compile "+loop"
	dw A.1(pluslp)	;   under rtn stk
	dw 2	
	dw compile	
	dw return	;   return
plu1:	dw stlod	; if false,
	db 5,'+LOOP'	;   print error msg
	dw wordq	
	dw return	;   return

; repeat (compiler word): "Repeat" compiles
;	the address of the executable code
;	for the word "repeat" under the
;	return stack. Sets TOS=5 as an
;	indication that it has occurred.

repeat1: sep call	; call 8th
	dw A.0(repeat)	; compile "repeat"
	dw A.1(repeat)	;   under rtn stk
	dw 2
	dw compile
	dw 5
	dw return	; return to next

; until (compiler word): "Until" compiles
;	the address of the executable
;	code for the word "until" under
;	the return stack. Before
;	compilation, tests for TOS=5,
;	which indicates that a
;	corresponding "until" exists.
;
until1:	sep call	; call 8th
	dw dup		; test for "repeat"
	dw 5		;   code=5 to be
	dw equal	;   sure "repeat"
	dw br0		;   preceeded "until"
	dw unt1		; if true,
	dw drop
	dw A.0(until)	;   compile "until"
	dw A.1(until)	;   under rtn stk
	dw 2
	dw compile
	dw return	;   return to next
unt1:	dw stlod	; if false,
	db 5,'UNTIL'	;   print error msg
	dw wordq
	dw return	;   return

; abort (compiler word): Used in compiler
;	mode to	abort the compilation of
;	the word being compiled. Erases
;	the word and clears the stacks.
;
abort2:	sep call	; call 8th
	dw curntat	; get last word
	dw dat		;   and erase it
	dw stat
	dw erase
	dw return	; return to next

; : (colon): ":" begins the definition
;	of a new word to be added to the
;	current vocabulary. The name will
;	be the next word following the
;	":" in the buffer. Each succesive
;	word is then compiled into the new
;	definition until a ";" is found.
;
;	The new word is entered into the
;	current vocabulary by moving the
;	stack down, placing the name in
;	the hole created, and linking it
;	to the current vocabulary. The
;	executable code for the new word
;	is inserted by moving the return
;	stack up and placing the code
;	in the resulting hole.
;
colon:	sep call	; call 8TH
	dw create	; create dict. def.
	dw 0d4h		; push 'call 8TH'
	dw comma	;   under rtn stack
	dw 1		; start loop to look
	dw repeat	;   for words
	dw getword	; get next word
	dw compiler	; search in
	dw dat		;   compiler vocab
	dw find
	dw br0
	dw comp1	; execute if found
	dw dat		;   in compiler vocab
	dw execute
	dw br		; else search the
	dw comp5	;   current and
comp1:	dw curntat	;   context vocab
	dw dat
	dw find
	dw dupnz
	dw not1
	dw br0
	dw comp2
	dw context
	dw dat
	dw dat
	dw find
comp2:	dw br0		; if found, leave
	dw comp3	;   executable addr.
	dw dat		;   under rtn stk
	dw 2
	dw compile
	dw br		; else try converting
	dw comp5	;   to a number
comp3:	dw numberq
	dw br0		; if not a number
	dw comp5	;   type "word ?"
	dw 0	 	;   compile the number
	dw 2
	dw compile
comp5:	dw memful	; if memory is full,
	dw br0
	dw comp6	;   erase word being
	dw curntat	;   compiled
	dw dat
	dw stat
	dw erase
comp6:	dw dupnz
	dw not1
	dw until	;   process next word
	dw return	;   return to next
	page
;--------------------------------------
;	     Dictionary
;	Names and Addresses
;--------------------------------------
;
; This part of the dictionary contains
; all the words initially part of 8TH
; following "clear" or a hardware clear.
; Each entry consists of the word itself
; (in string format: length byte first)
; followed by the address where the
; executable code for the word is
; located.  Entries and deletions are
; made at the low-address end.  The
; last entry is always 0 (null string)
; followed by address=0000.
;
vcomp:	equ  $		; compiler vocab
	db 5,'ABORT'
	dw abort2
	
	db 5,'UNTIL'
	dw until1

	db 6,'REPEAT'
	dw repeat1
	
	db 5,'+LOOP'
	dw pluslp1

	db 4,'LOOP'
	dw loop1

	db 2,'DO'
	dw do1

	db 5,'ENDIF'
	dw endif

	db 4,'ELSE'
	dw else

	db 2,'IF'
	dw if

	db 1,'"'
	dw quote1

	db 1,';'
	dw semicoln
	
	db 0
	dw 0
	
v8th:	equ $		; 8TH vocab

	db 1,':'
	dw colon

	db 8,'VARIABLE'
	dw varible

	db 5,'ALLOT'
	dw allot

	db 1,','
	dw comma

	db 7,'COMPILE'
	dw compile

	db 6,'CREATE'
	dw create

	db 6,'FORGET'
	dw forget

	db 5,'ERASE'
	dw erase

	db 3,'D1+'
	dw dinc

	db 5,'DSIZE'
	dw datasize

	db 8,'COMPILER'
	dw compiler

	db 3,'8TH'
	dw eighth

	db 4,'USER'
	dw user

	db 4,'SIZE'
	dw size

	db 1,'"'
	dw quote

	db 2,'BS'
	dw backspce

	db 3,'CAN'
	dw cancel

	db 3,'HEX'
	dw hex

	db 7,'DECIMAL'
	dw decimal

	db 1,'.'
	dw dot

	db 2,'D.'
	dw ddot

	db 9,'INTERPRET'
	dw intrpret

        db 1,"'"
	dw tick

	db 7,'NUMBER?'
	dw numberq

	db 5,'WORD?'
	dw wordq

	db 7,'GETWORD'
	dw getword

	db 1,'#'
	dw tonumber

	db 4,'DDUP'
	dw ddup

	db 6,'NUMBER'
	dw number

	db 7,'DNUMBER'
	dw dnumber

	db 6,'EXPECT'
	dw expect

	db 7,'GETAKEY'
	dw getakey

	db 5,'CHECK'
	dw check

	db 6,'MEMFUL'
	dw memful

	db 2,'CR'
	dw cr

	db 4,'TYPE'
	dw type

	db 4,'EMIT'
	dw emit

	db 6,'OUTPUT'
	dw output

	db 5,'INPUT'
	dw input

	db 6,'CURNT@'
	dw curntat

	db 5,'BASE@'
	dw basat

	db 6,'$ABORT'
	dw stabort

	db 8,'LASTKEY#'
	dw lastkeyn

	db 4,'BAUD'
	dw baud

	db 7,'#CYCLES'
	dw nocycles

	db 4,'ECHO'
	dw echo1

	db 4,'CAPS'
	dw caps

	db 6,'BUFFER'
	dw buffer

	db 9,'DELIMITER'
	dw delmiter

	db 10,'TERMINATOR'
	dw termnatr

	db 4,'BASE'
	dw base

	db 1,'N'
	dw n

	db 7,'CONTEXT'
	dw context

	db 7,'CURRENT'
	dw current

	db 3,'DMA'
	dw dma2

	db 3,'VAT'
	dw vat

	db 9,'INTERRUPT'
	dw intrupt1

	db 4,'CALL'
	dw call2

	db 4,'NEXT'
	dw next2

	db 3,'1ST'
	dw first

	db 4,'RSP0'
	dw rsp0

	db 3,'SP0'
	dw sp0

	db 6,'MINRAM'
	dw minram

	db 6,'MAXRAM'
	dw maxram

	db 3,'R>D'
	dw rtnmove

	db 5,'DSWAP'
	dw dswap

	db 2,'D<'
	dw dless

	db 3,'D>R'
	dw datamove

	db 4,'RSP@'
	dw rspat

	db 3,'SP@'
	dw spat

	db 2,'D!'
	dw dwri

	db 2,'D@'
	dw dat

	db 1,'/'
	dw divide

        db 2,"+'"
	dw addc

	db 5,'DIGIT'
	dw digit

	db 4,'$DUP'
	dw stdup

        db 2,"*'"
	dw multc

	db 1,'*'
	dw mult

	db 7,'EXECUTE'
	dw execute

	db 2,'LF'
	dw lf

	db 5,'DOVER'
	dw dover

	db 4,'FIND'
	dw find

	db 4,'WORD'
	dw word

	db 2,'D-'
	dw dsub

	db 5,'$DROP'
	dw stdr

	db 2,'$@'
	dw stat

	db 2,'$L'
	dw stlod

	db 2,'$!'
	dw stwri

	db 2,'D+'
	dw dadd

	db 6,'APPEND'
	dw append

	db 5,'DUPNZ'
	dw dupnz

	db 4,'DMA!'
	dw dmawri

	db 2,'Q!'
	dw qwri

	db 4,'KEYS'
	dw keys

	db 5,'EMITS'
	dw emits

	db 2,'V@'
	dw varat

	db 1,'!'
	dw excla

	db 1,'@'
	dw at

	db 3,'IE@'
	dw ieat

	db 2,'IE'
	dw ie

	db 2,'BR'
	dw br

	db 3,'BR0'
	dw br0

	db 4,'EXIT'
	dw exit

	db 1,'I'
	dw i

        db 7,'(+LOOP)'
	dw pluslp

        db 6,'(LOOP)'
	dw loop

        db 4,'(DO)'
	dw do

        db 7,'(UNTIL)'
	dw until

        db 8,'(REPEAT)'
	dw repeat

	db 1,'<'
	dw less

	db 1,'>'
	dw greater

	db 1,'='
	dw equal

	db 3,'NOT'
	dw not1

	db 3,'XOR'
	dw xor1

	db 2,'OR'
	dw or1

	db 3,'AND'
	dw and1

	db 1,'-'
	dw subtr

	db 1,'+'
	dw add

	db 3,'ROT'
	dw rot

	db 4,'SWAP'
	dw swap

	db 4,'OVER'
	dw over

	db 4,'DROP'
	dw drop

	db 5,'DDROP'
	dw ddrop

	db 3,'DUP'
	dw dup

	db 6,'RETURN'
	dw return

        db 7,"'ABORT'"
	dw abort

	db 5,'ABORT'
	dw abort

	db 5,'RESET'
	dw reset

	db 5,'CLEAR'
	dw clear

	db 3,'CLR'
	dw clr

	db 0
	dw 0		; end of 8th vocab

	page
;-------------------------------------------
;	INITIAL VALUES FOR VARIABLES	 
;-------------------------------------------
;
; initable: Table of initial values for
;	dictionary variables and CPU
;	registers. "Initable" may reside
;	in ROM, so is moved into RAM by a
;	hardware clear or "CLEAR". The
;	label "initable" is at the variable
;	base address, and so is at the end
;	of the table (8TH variables have a
;	negative offset from the base
;	address).
;
; INITIAL VALUE		VARIABLE NAME	REGISTER
; -------------		-------------	--------
endinit: org $
	db 0		; not a variable: marks
	dw 0000h	;   end of user vocab
	dw vcomp	; compile vocabulary
v8th1:	dw v8th		; 8TH vocabulary
	dw 0000h	; user vocabulary
	dw 0000h	; size
	db 08h		; backspace
	db 18h		; cancel
	dw emits	; output
	dw keys		; input
	db 00h		; lastkey#
	db 62h		; baud (300 @ 2MHz)
	db 0ffh		; #cycles
	db 01h		; echo flag
	db 01h		; caps flag
bufr:	db 79		; buffer
	db 13,10	; sign-on message
	db '    8TH, REVISION 1.0',13,10
	db '    COPYRIGHT 1981 BY',13,10
	db 'TECHNICAL MICRO SYSTEMS, INC'
	db '.', 13, 10
	org bufr+1+80
	db 32		; delimiter (space)
	db 13		; terminator (CR)
	db 10		; base	 
	db 0,0		; n
	dw 01f85h	; context vocalulary
	dw 01f87h	; current vocabulary
	dw 0ffffh	; DMA		(R0)
	dw varat	; variable	(R3)   
	dw interupt	; interrupt	(R1)
	dw call1	; call		(R4)
	dw next1	; next		(R5)
	dw intrpret	; first		(R7)
	dw $+8		; rsp0		(R6)
			; sp0		(R2)
	dw 3fffh-initable+endinit-1
	dw $+4		; minram	(R8)
	db 3fh		; maxram	(R9)
initable: db 0ffh
	end
